//////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2003-2012 Daniel Vik
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//////////////////////////////////////////////////////////////////////////////

#include "Crc32.h"

// If INVERSE_CRC is set, the algorithm returns the inverse CRC. This is the
// normal behavior. Some implementations does not inverse the CRC. Uncomment
// the define if the latter is the desired behavior.

#define INVERSE_CRC

//////////////////////////////////////////////////////////////////////////////
// Macros used for patch CRC to ASCII. These macros allows characters
// A-Z, a-z, and 0-9 to be part of the patch string.
//////////////////////////////////////////////////////////////////////////////

#define TO_CHAR(a) ((a) < 10 ? (a) + '0' : (a) < 36 ? ((a) - 10 + 'A') : (a) - 36 + 'a')
#define IS_CHAR(a) (((a) >= '0' && (a) <= '9') || ((a) >= 'A' && (a) <= 'Z') || ((a) >= 'a' && (a) <= 'z'))


#ifdef INVERSE_CRC
#define INV(x) ~(x)
#else
#define INV(x) (x)
#endif

//////////////////////////////////////////////////////////////////////////////
// CRC lookup table and patch table
//////////////////////////////////////////////////////////////////////////////


UInt32 Crc32::crc32Polynomial;

UInt32 Crc32::crc32Table[256];

int Crc32::patchCrc32Table[256];

//////////////////////////////////////////////////////////////////////////////
// Storage for patch bytes
//////////////////////////////////////////////////////////////////////////////

UInt8 Crc32::patchBytes[4];
char  Crc32::patchChars[7];


//////////////////////////////////////////////////////////////////////////////
// Constructor. Initializes CRC object
//////////////////////////////////////////////////////////////////////////////

Crc32::Crc32() : crc_(INV(0))
{
    CalculateTables();
}


//////////////////////////////////////////////////////////////////////////////
// Destructor
//////////////////////////////////////////////////////////////////////////////

Crc32::~Crc32()
{
}


//////////////////////////////////////////////////////////////////////////////
// Calculate lookup tables for any polynomial. The default tables use the
// polynomial 0x04c11db7.
//////////////////////////////////////////////////////////////////////////////
void Crc32::CalculateTables(UInt32 polynomial)
{
    if (polynomial == crc32Polynomial) {
        return;
    }

    crc32Polynomial = polynomial;

    // Create CRC lookup table
    for(int i = 0; i < 256; i++) {
        crc32Table[i] = Reflect(i, 8) << 24;
        for (int j = 0; j < 8; j++) {
            crc32Table[i] = (crc32Table[i] << 1) ^ (crc32Table[i] & (1 << 31) ?
                            polynomial : 0);
        }
        crc32Table[i] = Reflect(crc32Table[i], 32);
    }

    // Create patch CRC lookup table
    for (int j = 0; j < 256; j++) {
        for (int i = 0; i < 256; i++) {
            if ((j ^ (crc32Table[i] >> 24)) == 0) {
                patchCrc32Table[j] = i;
                break;
            }
        }
    }
}


//////////////////////////////////////////////////////////////////////////////
// Helper method to reflect a value, i.e. swap positions of bits 0 and 7,
// 1 and 6, etc.
//////////////////////////////////////////////////////////////////////////////
UInt32 Crc32::Reflect(UInt32 value, int count)
{
    for(int i = 0; i < count / 2; i++) {
        UInt32 tmp = ((value >> i) ^ (value >> (count - 1 - i))) & 1;
        value ^= ((tmp << i) | (tmp << (count - 1 - i)));
    }
    return value;
}


//////////////////////////////////////////////////////////////////////////////
// Appends data to the CRC object. The CRC hash is updated with CRC
// calculations of the provided buffer.
//////////////////////////////////////////////////////////////////////////////

void Crc32::append(const void* buffer, UInt32 size)
{
    const UInt8* ptr = (const UInt8*)buffer;
    crc_ = INV(crc_);

    while (size--) {
    	crc_ = (crc_ >> 8) ^ crc32Table[*ptr++ ^ (crc_ & 0xff)];
    }

    crc_ = INV(crc_);
}


//////////////////////////////////////////////////////////////////////////////
// Appends a byte to the CRC object. The CRC hash is updated with CRC
// calculations of the provided value.
//////////////////////////////////////////////////////////////////////////////

void Crc32::append(UInt8 value)
{
    crc_ = INV(crc_);
    crc_ = (crc_ >> 8) ^ crc32Table[value ^ (crc_ & 0xff)];
    crc_ = INV(crc_);
}


//////////////////////////////////////////////////////////////////////////////
// Rewinds the CRC based on the provided buffer. The CRC hash is updated with
// to reflect the CRC that would have been before the buffer was appended.
//////////////////////////////////////////////////////////////////////////////

void Crc32::rewind(const void* buffer, UInt32 size)
{
    const UInt8* ptr = (const UInt8*)buffer;
    crc_ = INV(crc_);

    for (int i = size - 1; i >= 0; --i) {
        for (int j = 0; j < 256; ++j) {
            if ((crc32Table[j] >> 24) == (crc_ >> 24)) {
                crc_ = ((crc_ ^ crc32Table[j]) << 8) | (j ^ ptr[i]);
                break;
            }
        }
    }

    crc_ = INV(crc_);
}


//////////////////////////////////////////////////////////////////////////////
// Rewinds the CRC based on the provided value. The CRC hash is updated with
// to reflect the CRC that would have been before the value was appended.
//////////////////////////////////////////////////////////////////////////////

void Crc32::rewind(UInt8 value)
{
    crc_ = INV(crc_);

    for (int j = 0; j < 256; ++j) {
        if ((crc32Table[j] >> 24) == (crc_ >> 24)) {
            crc_ = ((crc_ ^ crc32Table[j]) << 8) | (j ^ value);
            break;
        }
    }

    crc_ = INV(crc_);
}


//////////////////////////////////////////////////////////////////////////////
// Calculates the CRC at an earlier offset in a buffer, based on the buffer
// itself and the CRC at the end of the buffer, returning all possible CRCs
// at the beginning of the buffer. Most CRC polynomials gives one unique CRC,
// but the method finds all possible ones.
//////////////////////////////////////////////////////////////////////////////

std::list<UInt32> Crc32::Rewind(void* buffer, UInt32 size, UInt32 endCrc) {
    std::list<UInt32> crcs;
    std::stack<CrcMatch*> stack;

    // Push last octet in the buffer to seed the loop.
    stack.push(new CrcMatch(buffer, size, INV(endCrc)));

    while (!stack.empty()) {
	CrcMatch* node = stack.top();
        stack.pop();
        if (node->offset == 0) {
            // Reached the beginning of the buffer; save the CRC.
            crcs.push_back(INV(node->crc));
        }
        else {
            for (int i = 0; i < 256; ++i) {
                if ((crc32Table[i] >> 24) == (node->crc >> 24)) {
                    UInt32 prevCrc = ((node->crc ^ crc32Table[i]) << 8) |
                                     (i ^ node->ptr[node->offset - 1]);
                    stack.push(new CrcMatch(node->ptr, node->offset - 1,
                                            prevCrc));
                }
            }
        }
        delete node;
    }
    return crcs;
}


//////////////////////////////////////////////////////////////////////////////
// Finds CRC patch bytes. The method returns four bytes that can be appended
// to the CRC in order to get the desired CRC
//////////////////////////////////////////////////////////////////////////////

const UInt8* Crc32::findPatch(UInt32 desiredCrc) const
{
    int crcIdx[4];

    UInt32 iterCrc = INV(desiredCrc);
    for (int j = 3; j >= 0; j--) {
        crcIdx[j] = patchCrc32Table[iterCrc >> 24];
        iterCrc = (iterCrc ^ crc32Table[crcIdx[j]]) << 8;
    }

    UInt32 crc = INV(crc_);
    for (int j = 0; j < 4; j++) {
        patchBytes[j] = (crc ^ crcIdx[j]) & 0xff;
    	crc = (crc >> 8) ^ crc32Table[patchBytes[j] ^ (crc & 0xff)];
    }

    return patchBytes;
}

//////////////////////////////////////////////////////////////////////////////
// Finds CRC patch bytes. The method returns six ASCII characters
// (a-z,A-Z,0-9) that can be appended to the CRC in order to get the desired
// CRC.
//////////////////////////////////////////////////////////////////////////////

const char* Crc32::findPatchAscii(UInt32 desiredCrc) const
{
    Crc32 crc32;

    for (int a = 0; a < 62; a++) {
        for (int b = 0; b < 62; b++) {
            crc32.set(crc_);
            crc32.append(TO_CHAR(a));
            crc32.append(TO_CHAR(b));
            const UInt8* patch = crc32.findPatch(desiredCrc);

            // If the patch CRC has valid ASCII characters, we found a solution.
            if (IS_CHAR(patch[0]) && IS_CHAR(patch[1]) &&
                IS_CHAR(patch[2]) && IS_CHAR(patch[3])) {
                patchChars[0] = (char)TO_CHAR(a);
                patchChars[1] = (char)TO_CHAR(b);
                patchChars[2] = (char)patch[0];
                patchChars[3] = (char)patch[1];
                patchChars[4] = (char)patch[2];
                patchChars[5] = (char)patch[3];
                patchChars[6] = 0;
                return patchChars;
            }
        }
    }

    return "";
}


//////////////////////////////////////////////////////////////////////////////
// Sets a new CRC hash value
//////////////////////////////////////////////////////////////////////////////

void Crc32::set(UInt32 crc)
{
    crc_ = crc;
}


//////////////////////////////////////////////////////////////////////////////
// Gets current CRC hash value
//////////////////////////////////////////////////////////////////////////////

UInt32 Crc32::get() const
{
    return crc_;
}
